package IPC;

import java.net.DatagramPacket;
import java.net.DatagramSocket;
import java.io.ByteArrayOutputStream;
import java.io.DataOutputStream;
import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.IOException;

public class Message {

    //
    // Maximum size of data buffer (including header)
    //
    final static int maxBufSize   = 1000;

    //
    // Specify layout of message header
    //
    final static int hdrOffType         = 0;
    final static int hdrOffServerEpoch  = 4;
    final static int hdrOffSeqNum       = 8;
    final static int hdrSizeBase        = 12;

    //
    // Local data
    //
    DatagramPacket        packet;
    protected byte[]      buffer;
    int                   headerSize;
    protected int                   dataCursor;

    //
    // Streams used for to marshal and unmarshal
    //
    protected ByteArrayOutputStream baos;
    protected DataOutputStream      dos;
    protected ByteArrayInputStream  bais;
    protected DataInputStream       dis;
    protected ByteArrayInputStream  bais_hdr;
    protected DataInputStream       dis_hdr;

    //
    // Constructor
    //
    public
    Message ()
    {
	buffer     = new byte [maxBufSize];
	packet     = new DatagramPacket        (buffer, maxBufSize);
	baos       = new ByteArrayOutputStream ();
	dos        = new DataOutputStream      (baos);
	bais_hdr   = new ByteArrayInputStream  (buffer, 0, getHeaderSize ());
	dis_hdr    = new DataInputStream       (bais_hdr);
	bais       = new ByteArrayInputStream  (buffer, 0, getHeaderSize ());
	dis        = new DataInputStream       (bais);
	headerSize = getHeaderSize             ();
	reset                                  ();
    }

    protected DatagramPacket
    getPacket()
    {
      return packet;
    }
    
    protected byte[]
    getBuffer()
    {
      return buffer;
    }

    public void
    copyFrom (Message that)
    {
        packet     = that.packet;
	buffer     = that.buffer;
	dataCursor = that.dataCursor;
	baos       = that.baos;
	dos        = that.dos;
	bais_hdr   = new ByteArrayInputStream (buffer, 0, getHeaderSize());
	dis_hdr    = new DataInputStream      (bais_hdr);
	bais       = new ByteArrayInputStream (buffer, 0, packet.getLength());
	dis        = new DataInputStream      (bais);
	reset ();
    }

    //
    // Protected method to get header size
    // SPECIALIZED BY SUBCLASSES
    //
    protected int
    getHeaderSize ()
    {
	return hdrSizeBase;
    }

    //
    // Protected method to send a message
    //
    public void
    send (DatagramSocket socket)
        throws IOException
    {
        packet.setLength (dataCursor + getHeaderSize ());
	socket.send      (packet);
    }

    //
    // Protected method to receive a message
    //
    public void
    receive (DatagramSocket socket)
        throws IOException
    {
    	packet.setLength		(maxBufSize);  // work around Java bug
        socket.receive                  (packet);
        bais = new ByteArrayInputStream (buffer, 0, packet.getLength());
	dis  = new DataInputStream      (bais);
	reset                           ();
    }

    //
    // Private method to read integer from specified position in message header
    //
    protected int
    getIntAt (int offset)
	throws MarshalException
    {
        if ( (offset+4) > packet.getLength() )
    	    throw new MarshalException ();
	bais_hdr.reset ();
	try {
	    dis_hdr.skipBytes (offset);
	    return dis_hdr.readInt ();
        } catch (IOException ie) {
	    throw new MarshalException ();
        }
    }

    //
    // Private method to write integer at specified position in message header
    //
    protected void
    setIntAt (int offset, int num)
    {
        byte[] bytes;

	baos.reset ();
	try {
	    dos.writeInt (num);
        } catch (IOException ie) {
	    throw new AssertException ("Unexpection IO Exception in setIntAt");
        }
        bytes = baos.toByteArray();
	for (int i=0; i<baos.size(); i++) {
	    buffer [offset+i] = bytes[i];
        }
    }
    
    //
    // Private method to add content of data-output stream to message data
    //
    protected void
    addFromDOS ()
        throws MessageFullException
    {
	int    bufStart;
	int    bytesSize;
	byte[] bytes;

	//
	// compute buffer position of new data
	//
	bufStart  = dataCursor + getHeaderSize ();
	bytes     = baos.toByteArray ();
	bytesSize = baos.size        ();
	
	//
	// check for buffer overrun
	//
	if (bufStart + bytesSize > maxBufSize) {
	    throw new MessageFullException ();
	}
	
	//
	// copy binary form of num into buffer
	//
	for (int i=0; i<bytesSize; i++) {
	    buffer [bufStart+i] = bytes[i];
	}
	
	//
	// update buffer-position cursor 
	//
	dataCursor += baos.size();
	
	//
	// update packet size
	//
    }


    
    //
    // Methods for reading message header
    //
   
    ComHandle
    getComHandle ()
    {
	int epoch;
		
	try {
	    epoch = getServerEpoch();
	} catch (IPC.MarshalException e) {
	    epoch = 0;
	}
	
	return new ComHandle (packet.getAddress(), packet.getPort(), epoch);
    }

    int
    getType ()
        throws MarshalException
    {
        return getIntAt (hdrOffType);
    }

    int
    getSeqNum ()
        throws MarshalException
    {
	return getIntAt (hdrOffSeqNum);
    }
    
    int
    getServerEpoch ()
    	throws MarshalException
    {
    	return getIntAt (hdrOffServerEpoch);
    }
    
    //
    // Methods for setting message header
    //

    public void 
    setComHandle (ComHandle dest)
    {
	packet.setAddress     (dest.getAddress());
	packet.setPort        (dest.getPort());
	setServerEpoch        (dest.getEpoch());
    }

    void
    setType (int type)
    {
        setIntAt (hdrOffType, type);
    }

    void
    setSeqNum (int seqNum)
    {
        setIntAt (hdrOffSeqNum, seqNum);
    }
    
    void
    setServerEpoch (int serverEpoch)
    {
    	setIntAt (hdrOffServerEpoch, serverEpoch);
    }

    //
    // Public Methods (Marshaling)
    //

    //
    // Reset current data position to beginning (for addXXX and getXXX)
    //
    public void
    reset ()
    {
    	//
    	// reset buffer position used by addInt()
    	//
	dataCursor = 0;
	
	//
	// reset buffer position used by getInt()
	//
	bais.reset ();
	try {
	    dis.skipBytes (getHeaderSize());
        } catch (IOException ie) {
	    throw new AssertException ("Unexpected IOException in reset");
        }
    }
    
    //
    // Add integer to message data (at current position and advance position)
    //
    public void
    addArgInt (int num)
        throws MessageFullException
    {
    
	//
	// write number to data output stream
	//
        try {
            baos.reset ();
            dos.writeInt (MessageDatatype.INT);
            dos.writeInt (num);
        } catch (IOException ie)  {
            throw new AssertException ("Unexpected IO Exception in addArgInt");
        }
        
        //
        // now copy bits in dos to message data
        //
        addFromDOS ();
    }

    //
    // Read integer from message data (at current position and advance position)
    //
    public int
    getArgInt ()
        throws MarshalException
    {
	//
	// read int from current position in data input stream (msg buffer)
	//
	try {
	    if (dis.readInt() != MessageDatatype.INT)
	        throw new MarshalException ();
	    return dis.readInt ();
	} catch (IOException ie) {
	    throw new MarshalException ();
	}
    }
    
    public void
    addArgString (String str)
        throws MessageFullException
    {
    	try {
    	    baos.reset     ();
    	    dos.writeInt   (MessageDatatype.STRING);
    	    dos.writeInt   (str.length());
    	    dos.writeBytes (str);
    	} catch (IOException ie) {
    	    throw new AssertException ("Unexpected IO Exception in addArgInt");
    	}
    	addFromDOS ();
    }
    
    public String
    getArgString ()
        throws MarshalException
    {
    	byte data [];
    	int  len;
    	
    	try {
    	    if (dis.readInt() != MessageDatatype.STRING)
    	        throw new MarshalException ();
    	    len  = dis.readInt ();
    	    data = new byte [len];
    	    dis.readFully (data, 0, len);
    	    return new String(data, 0, len);
    	} catch (IOException ie) {
    	    throw new MarshalException ();
    	}
    }
}


