package IPC;

import java.io.IOException;
import java.net.DatagramSocket;
import java.net.SocketException;
import java.net.InetAddress;
import java.net.UnknownHostException;


public abstract class Server  {

    //
    // Private Data
    //
    int             epoch;
    ClientList      clientList;
    DatagramSocket  comSocket;
    boolean         up;
    Object          upCond;
    PendingRequests pendingRequests;

    //
    // Constructor
    //
    public
    Server (int epoch) 
    	throws IOException
    {
        clientList      = new ClientList ();
        up              = false;
	upCond          = new Object ();
	comSocket       = new DatagramSocket ();
	this.epoch      = epoch;
	pendingRequests = new PendingRequests (comSocket);
    }

    //
    // Return a new server request for this reqMsg;
    // IMPLEMENTED BY SUBCLASSES
    //
    protected abstract ServerRequest
    getNewServerRequest (Message reqMsg);
    
    //
    // Return a new request message
    // IMPLEMENTED BY SUBCLASSES
    //
    protected abstract Message
    getNewRequestMessage ();

    //
    // Called by ResponseMessage to send response to client
    //
    void
    sendResponse (Message respMsg)
    {
	ComHandle client;
	int       seqNum;

	try {
	    client = respMsg.getComHandle ();
	    seqNum = respMsg.getSeqNum    ();          
	    clientList.saveResponse       (client, seqNum, respMsg);
	    pendingRequests.cancel        (client, seqNum);
	    respMsg.send                  (comSocket);
        } catch (IOException ie) {
	    throw new AssertException ("Unexpected IO Exception sending response");
	} catch (MarshalException me) {
	    throw new AssertException ("Response message garbled");
	}
    }

    //
    // Return a ComHandle for this Server
    //
    public ComHandle
    getComHandle ()
    {
	try {
	    return new ComHandle (InetAddress.getLocalHost (),
			          comSocket.getLocalPort   (),
			          epoch);
        } catch (UnknownHostException uhe) {
	    throw new AssertException ("Bad local host in server.getComHandle");
        }
    }

    //
    // Stop the receiver thread
    //
    public void
    shutdown ()
    {
	Message msg;

        up = false;
        //
        // send bogus message to server to unblock receive thread
        //
	msg = new Message  ();
        msg.setComHandle   (getComHandle ());
        msg.setType        (MessageType.SHUTDOWN_REQUEST);
	msg.setSeqNum      (-1);
	msg.setServerEpoch (epoch);
	try {
	    msg.send (comSocket);
	} catch (IOException ie) {
	    System.out.println 
		("Server shutdown send threw a " + ie.getClass() +
		 "\n    with message: " + ie.getMessage());
	}
    }
    
    //
    // loop waiting for requests (started in new thread by Start())
    //
    public void
    start ()
    {
	Message         reqMsg;
	Message         respMsg;
        ComHandle       clientComHandle;
	int             reqType;
	int             reqSeqNum;
	int             reqEpoch;
	RequestStatus   requestStatus;

	synchronized (upCond) {
	    up = true;
	    upCond.notifyAll();
	}

        System.out.println ("IPC.Server started");

	while (up) {
	    
            //
            // blocking wait for next request message
            //
            
	    try {
	    
	    	//
	    	// create a new message buffer and block in receive
	    	//
	    	reqMsg = getNewRequestMessage ();
	        reqMsg.receive                (comSocket);
	        
	    } catch (IOException ie) {
	    
		//
		// unexpected receive error; die
                //
                throw new AssertException ("Server receive threw a " + ie.getClass() +
			                   "\n    with message: " + ie.getMessage());
	    }
	    
	    //
	    // deal with request message
	    //
	    
	    try {
	    
	    	//
	    	// read the request header
	    	//
	        clientComHandle = reqMsg.getComHandle   ();
	        reqType         = reqMsg.getType        ();
	        reqSeqNum       = reqMsg.getSeqNum      ();
	    	reqEpoch        = reqMsg.getServerEpoch ();

		//
		// respond to ARE_YOU_ALIVE messages
		//
		if (reqType == MessageType.ARE_YOU_ALIVE) {
		    respMsg = new Message ();
		    respMsg.setComHandle (clientComHandle);
		    respMsg.setType      (MessageType.I_AM_ALIVE);
		}
	    	
	    	//
	    	// make sure its a request - ignore otherwise
	    	//
	    	else if (reqType != MessageType.REQUEST) {
	    	    respMsg = null;
	        }
	    	
	    	//
	    	// check for stale epoch - error
	    	//
	    	else if (reqEpoch != epoch) {
	    	    respMsg = new Message ();
	    	    respMsg.setType       (MessageType.STALE_EPOCH);
	    	}

	    	//
	    	// check for old seqnum and otherwise fork thread to handle request
	    	//
		else {

		    requestStatus = clientList.newRequest (clientComHandle, reqSeqNum);

		    if (requestStatus != null) {

		        switch (requestStatus.status) {

			    case RequestStatusType.ACTIVE:
			        respMsg = new Message  ();
				respMsg.setType        (MessageType.I_AM_ALIVE);
				pendingRequests.cancel (clientComHandle, reqSeqNum);
				break;

			    case RequestStatusType.SAVED:
				respMsg = requestStatus.respMsg;
				break;

			    case RequestStatusType.STALE:
			        respMsg = new Message  ();
			        respMsg.setType        (MessageType.BAD_SEQNUM_ERR);
			        break;

			    default:
			        throw new AssertException ("RequestStatusType");
			}
		    } else {

			respMsg = null;
			pendingRequests.register (clientComHandle, reqSeqNum);
			(new Thread ( getNewServerRequest (reqMsg)) ).start ();
		    }
	    	}	    
	    	
	    	//
	    	// finally, send response message if we have one 
	    	//
	    	if (respMsg != null) {
		    respMsg.setComHandle   (clientComHandle);
		    respMsg.setSeqNum      (reqSeqNum);
		    respMsg.setServerEpoch (epoch);
	    	    try {
			respMsg.send (comSocket);
	    	    } catch (IOException ie) {
	    	        throw new AssertException ("Unexpected IOException responding");
	    	    } finally {
	    	        respMsg = null;
	    	    }
	    	}

	    } catch (MarshalException me) {
	    
	        //
	        // garbled message; ignore the message (no response)
	        //
	        respMsg = null;	        
            }
	}
    }
}
